# -*- coding: utf-8 -*-
"""PyMzn is a Python library that wraps and enhances the MiniZinc tools for
modelling and solving constraint programs. It is built on top of the MiniZinc
toolkit and provides a number of off-the-shelf functions to readily solve
problems encoded in MiniZinc and parse the solutions into Python objects."""

import ast
import logging

from . import config
from . import dzn
from .dzn import *
from . import mzn
from .mzn import *
from .mzn import templates

__version__ = '0.16.12'
__all__ = ['debug', 'config']
__all__.extend(dzn.__all__)
__all__.extend(mzn.__all__)

# TODO: update python2 branch

_debug_handler = None
_pymzn_logger = logging.getLogger(__name__)
_pymzn_logger.addHandler(logging.NullHandler())

def debug(dbg=True):
    """Enables or disables debugging messages on the standard output."""
    global _debug_handler
    if dbg and _debug_handler is None:
        _debug_handler = logging.StreamHandler()
        _pymzn_logger.addHandler(_debug_handler)
        _pymzn_logger.setLevel(logging.DEBUG)
    elif not dbg and _debug_handler is not None:
        _pymzn_logger.removeHandler(_debug_handler)
        _debug_handler = None
        _pymzn_logger.setLevel(logging.WARNING)


def main():
    import argparse
    from textwrap import dedent

    def _minizinc(**_args):
        print(minizinc(_args['mzn'], *_args['dzn_files'],
            **{k: v for k, v in _args.items() if k not in ['mzn', 'dzn_files']}))

    def _config(key, value=None, **__):
        if value is None:
            print('{} : {}'.format(key, config.get(key)))
        else:
            config.set(key, value)
            config.dump()

    desc = dedent('''\
        PyMzn is a Python library that wraps and enhances the MiniZinc tools for
        modelling and solving constraint programs. It is built on top of the
        MiniZinc toolkit and provides a number of off-the-shelf functions to
        readily solve problems encoded in MiniZinc and parse the solutions into
        Python objects.
    ''')

    fmt = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(description=desc, formatter_class=fmt)
    parser.add_argument(
        '--version', action='version',
        version='PyMzn version: {}'.format(__version__)
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='display informative messages on standard output'
    )

    # minizinc
    parser.add_argument(
        'mzn', help='the mzn file to solve'
    )
    parser.add_argument(
        'dzn_files', nargs='*', help='additional dzn files'
    )
    parser.add_argument(
        '--data', type=ast.literal_eval, help='additional inline data'
    )
    parser.add_argument(
        '-a', '--all-solutions', action='store_true',
        help='wheter to return all solutions (if supported by the solver)'
    )
    parser.add_argument(
        '-S', '--solver', help='name of the solver'
    )
    parser.add_argument(
        '-s', '--solver-args', type=ast.literal_eval, default={},
        help='arguments to pass to the solver'
    )
    parser.add_argument(
        '-k', '--keep', action='store_true',
        help='whether to keep generated files'
    )
    parser.add_argument(
        '-I', '--include', dest='path', action='append',
        help='directory the standard library'
    )
    parser.set_defaults(func=_minizinc)

    # config
    subparsers = parser.add_subparsers()
    config_parser = subparsers.add_parser(
        'config', help='config pymzn variables'
    )
    config_parser.add_argument(
        'key', help='the property to get/set'
    )
    config_parser.add_argument(
        'value', nargs='?', help='the value(s) to set'
    )
    config_parser.set_defaults(func=_config)

    args = parser.parse_args()

    debug(args.verbose)
    args.func(**{**vars(args), **args.solver_args})


if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
"""

PyMzn can be configured with custom executable paths and other variables.
Configuration is done via the module ``pymzn.config``. For instance::

    import pymzn.config

    pymzn.config.set('mzn2fzn', 'path/to/mzn2fzn')
    pymzn.config.set('solns2out', 'path/to/solns2out')

The configurable properties used by PyMzn are the following:

 * **mzn2fzn**: Path to the *mzn2fzn* executable;
 * **solns2out**: Path to the *solns2out* executable;
 * **solver**: Solver instance to use when calling pymzn.minizinc;
 * **solver_args**: Arguments to pass to the solver when calling pymzn.minizinc;
 * **keep**: Overrides the keep flag of all minizinc and mzn2fzn calls;
 * **output_dir**: Set a default output directory for generated files;
 * **force_flatten**: Overrides the force_flatten flag of all minizinc calls;
 * **no_output_annotations**: Overrides the no_output_annotation flag of all
   minizinc calls;
 * **dzn_width**: The horizontal character limit for dzn files;
   This property is used to wrap long dzn statements when writing dzn files.
   This property is also used in the ``pymzn.minizinc`` function as a limit to
   decide whether to write the inline data into a file.

One can also set custom properties to be used for custom solvers.


Debug
-----

PyMzn can also be set to print debugging messages on standard output via::

    pymzn.debug()

This function is meant to be used in interactive sessions or in
applications that do not configure the ``logging`` library. If you configure the
``logging`` library in your application, then PyMzn will print logging messages
as well. The logging level in PyMzn is always ``DEBUG``. To disable debugging
messages you can then call::

    pymzn.debug(False)

"""
import os


_modified = False
_config = None
_defaults = {
    'mzn2fzn': 'mzn2fzn',
    'solns2out': 'solns2out',
    'dzn_width': 70
}


def _cfg_file():
    try:
        import appdirs
        return os.path.join(appdirs.user_config_dir(__name__), 'config.yml')
    except ImportError:
        return None


def get(key, default=None):
    """Get the value of a configuration variable.

    Parameters
    ----------
    key : str
        The key of the variable to retrieve.
    default
        The default value to return if the key does not exist.

    Returns
    -------
        The value associated to the key if the key exists, otherwise the default
        if provided.
    """
    global _config
    if _config is None:
        _config = {}
        try:
            import yaml
            cfg_file = _cfg_file()
            if cfg_file and os.path.isfile(cfg_file):
                with open(cfg_file) as f:
                    _config = yaml.load(f)
        except ImportError:
            pass
    if default is None:
        default = _defaults.get(key)
    return _config.get(key, default)


def set(key, value):
    """Set the value of configuration variable.

    Parameters
    ----------
    key : str
        The key of the variable to set.
    value
        The value to assign to the variable.
    """
    global _config
    global _modified
    if get(key) != value:
        _config[key] = value
        _modified = True


def dump():
    """Writes the changes to the configuration file."""
    global _config
    global _modified
    if _modified:
        try:
            import yaml
            cfg_file = _cfg_file()
            cfg_dir, __ = os.path.split(cfg_file)
            os.makedirs(cfg_dir, exist_ok=True)
            with open(cfg_file, 'w') as f:
                yaml.dump(_config, f)
            _modified = False
        except ImportError as err:
            raise RuntimeError(
                'Cannot dump the configuration settings to file. You need to '
                'install the necessary dependencies (pyyaml, appdirs).'
            ) from err




"""

PyMzn provides a set of functions to deal with dzn encoded strings and
files. Using these functions it is possible to serialize Python objects
into dzn format and vice-versa. For instance, the ``pymzn.dict2dzn`` function
converts an assignment of variables (provided as a dictionary) into dzn
format::

    pymzn.dict2dzn({'a': 2, 'b': {4, 6}, 'c': {1, 2, 3}, 'd': {3: 4.5, 4: 1.3}, 'e': [[1, 2], [3, 4], [5, 6]]})

The output is a list of dzn statements, as for the previous example::

    ['a = 2;', 'b = {4, 6};', 'c = 1..3;', 'd = array1d(3..4, [4.5, 1.3]);', 'e = array2d(1..3, 1..2, [1, 2, 3, 4, 5, 6];']

Optionally, you can pass the path to a dzn file where to write the
statements.

::

    pymzn.dict2dzn(data, fout='path/to/dzn')

The supported types of python objects are: \* Booleans \* Integers \*
Floats \* Sets \* Multi-dimensional arrays: \* lists are converted into
dzn arrays with index-set 1..len(list); \* dicts are converted into dzn
arrays with index-set equal to the key-set of the dict, provided that
the index-set is contiguous; \* nested combinations of the previous two,
provided that the children of every node have the same index-set. The
maximum depth is 6.

To evaluate back from dzn to python objects you can use the
``pymzn.dzn2dict`` function, which takes as input a dzn formatted
string or the path to a dzn file and returns the corresponding dictionary of
variable assignments. For instance, given the following dzn file::

    %% test.dzn %%

    a = 2;
    b = {4, 6};
    c = 1..3;
    d = array1d(3..4, [4.5, 1.3]);
    e = array2d(1..3, 1..2, [1, 2, 3, 4, 5, 6]);

Running the function::

    pymzn.dzn2dict('test.dzn')

will return::

    {'a': 2, 'b': {4, 6}, 'c': {1, 2, 3}, 'd': {3: 4.5, 4: 1.3}, 'e': [[1, 2], [3, 4], [5, 6]]}

which is identical to the object we serialized in the previous example.

In general, there is a one to one translation from python objects to dzn and
back, with the only exception of arrays with index-sets not based in 1. Arrays
and matrices based in 1 are translated into lists instead of dictionaries with
explicit keys. For instance::

    pymzn.dict2dzn({'a': {1: 2, 2: 4, 3: 6}})
    # returns: ['a = array1d(1..3, [2, 4, 6]);']

but when translating back the array, whose index-set is based in 1, will be
translated into a list::

    pymzn.dzn2dict('a = array1d(1..3, [2, 4, 6]);')
    # returns: {'a': [2, 4, 6]}

If you wish to avoid this behavior and get all arrays as dictionaries then you
can specify ``rebase_arrays=False`` as argument to the ``pymzn.dzn2dict``
function.

If, instead, you want to rebase also the arrays and matrices with different
index-sets you can use the ``pymzn.rebase_array`` function, which will discard
the index-sets in the dictionary representation of the array (matrix) and
transform it into a list (list of lists). For instance::

    pymzn.rebase_array({3: {2: 1, 3: 2, 4: 3}, 4: {1: 2, 2: 3}}, recursive=True)
    # returns: [[1, 2, 3], [2, 3]]

"""

from .marsh import *
from .eval import *

__all__ = ['dict2dzn', 'stmt2dzn', 'val2dzn', 'rebase_array', 'dzn2dict']




"""process.py

This package implements the `Process` class, a helper class that wraps
`subprocess.Popen` and make it safe to use both synchronously and
asynchronously.
"""

import io
import os

from threading import Thread, Lock
from time import monotonic as _time
from subprocess import Popen, PIPE, TimeoutExpired, CalledProcessError


__all__ = ['Process']


class Process:
    """Wrapper for an external process.

    Usable to run a synchronous process or to manage an asynchronous one.

    Parameters
    ----------
    args : [str]
        The command line arguments to execute. Same as for `subprocess.Popen`
        (with `shell=False`).

    Attributes
    ----------
    returncode : int or None
        The returncode of the process. While the process has not finished the
        returncode is None. After the process is finished, the returncode is
        retrieved as by `Popen.poll()`.
    stdout_data : str or bytes
        The content of the standard output of the process after the execution.
    stderr_data : str or bytes
        The content of the standard error of the process after the execution.
    expired : bool
        Whether the process was terminated because the timeout expired.
    interrupted : bool
        Whether the process was interrupted by a KeyboardInterruption.
    async : bool
        Whether the process was started asynchronously.
    started : bool
        Whether the process has started.
    alive : bool
        Whether the process is still running.
    completed : bool
        Whether the process was completed without errors.
    runtime : float
        The running time in seconds. If the process is alive, it marks the
        running time up to now. If instead the process is terminated, it marks
        the running time of the process up to its end.
    stdout
        The underlying output stream of the process.
    stderr
        The underlying error stream of the process.
    """
    def __init__(self, args):
        self.args = args
        self.returncode = None
        self.timeout = None
        self.stdout_data = None
        self.stderr_data = None
        self.expired = False
        self.interrupted = False
        self.async = False
        self._start = None
        self._end = None
        self._process = None
        self._process_lock = Lock()
        self._waiter_thread = None

    @property
    def started(self):
        return self._process is not None

    @property
    def alive(self):
        return self.started and self.returncode is None

    @property
    def completed(self):
        return self.returncode == 0

    @property
    def runtime(self):
        if not self._start:
            return 0.0
        end = self._end or _time()
        return end - self._start

    @property
    def stdout(self):
        return self._process.stdout

    @property
    def stderr(self):
        return self._process.stderr

    def _check_started(self):
        if self.started:
            raise RuntimeError('Process already started')

    def run(self, input=None, timeout=None):
        """Run the process synchronously.

        Parameters
        ----------
        input : str or bytes or None
            The content to for the input stream of the process.
        timeout : float or None
            The timeout for the process. None means no timeout.
        """
        self._check_started()
        popenkwargs = {'bufsize': 1, 'universal_newlines': True,
                       'stdin': PIPE, 'stdout': PIPE, 'stderr': PIPE}
        if os.name == 'nt':
            popenkwargs['shell'] = True
        self._process_lock.acquire()
        self.timeout = timeout
        self._start = _time()
        with Popen(self.args, **popenkwargs) as self._process:
            try:
                stdout, stderr = self._process.communicate(input, timeout)
                self.stdout_data, self.stderr_data = stdout, stderr
            except KeyboardInterrupt:
                self._process.kill()
                self._process.wait()
                self.interrupted = True
            except TimeoutExpired:
                self._process.kill()
                stdout, stderr = self._process.communicate()
                self.stdout_data, self.stderr_data = stdout, stderr
                self.expired = True
                raise TimeoutExpired(self.args, timeout, stdout, stderr)
            except:
                self._process.kill()
                self._process.wait()
                raise
            finally:
                self._end = _time()
                self.returncode = retcode = self._process.poll()
                self._process_lock.release()
                if not (self.expired or self.interrupted) and self.returncode:
                    stdout, stderr = self.stdout_data, self.stderr_data
                    raise CalledProcessError(retcode, self.args, stdout, stderr)
        return self

    def _wait(self):
        remaining = None
        if self.timeout:
            remaining = self.timeout + self._start - _time()
        try:
            self._process.wait(remaining)
        except KeyboardInterrupt:
            self._process_lock.acquire()
            self._process.kill()
            self._process.wait()
            self._process_lock.release()
            self.interrupted = True
        except TimeoutExpired:
            self._process_lock.acquire()
            self._process.kill()
            stdout, stderr = self._process.communicate()
            self._process_lock.release()
            self.stdout_data, self.stderr_data = stdout, stderr
            self.expired = True
            raise TimeoutExpired(self.args, timeout, stdout, stderr)
        except:
            self._process_lock.acquire()
            self._process.kill()
            self._process.wait()
            self._process_lock.release()
            raise
        finally:
            self._end = _time()
            self.returncode = retcode = self._process.poll()
            self._cleanup()
            if not (self.expired or self.interrupted) and self.returncode:
                stdout, stderr = self.stdout_data, self.stderr_data
                raise CalledProcessError(retcode, self.args, stdout, stderr)

    def start(self, stdin=None, timeout=None):
        """Starts the process asynchronously.

        Parameters
        ----------
        stdin : BufferedReader or None
            The stream to attach to the standard input of the process.
        timeout : float or None
            The timeout for the process. None means no timeout.
        """
        self._check_started()
        popenkwargs = {'bufsize': 0, 'universal_newlines': True,
                       'stdin': stdin, 'stdout': PIPE, 'stderr': PIPE}
        if os.name == 'nt':
            popenkwargs['shell'] = True
        self._process_lock.acquire()
        if not self.started:
            self.async = True
            self.timeout = timeout
            self._start = _time()
            self._process = Popen(self.args, **popenkwargs)
            self._waiter_thread = Thread(target=self._wait)
            self._waiter_thread.start()
            self._process_lock.release()
        else:
            self._process_lock.release()
            raise RuntimeError('Process already started')
        return self

    def _cleanup(self):
        if self._process.stdout:
            if not self._process.stdout.closed:
                self.stdout_data = self._process.stdout.read()
                self._process.stdout.close()
        if self._process.stderr:
            if not self._process.stderr.closed:
                self.stderr_data = self._process.stderr.read()
                self._process.stderr.close()
        if self._process.stdin:
            self._process.stdin.close()

    def stop(self):
        """Stops the process.

        Usable only if the process was started asynchronously.
        """
        if not self._process or not self.alive:
            return
        self._process_lock.acquire()
        self._process.kill()
        self._waiter_thread.join()
        self._process_lock.release()

    def readlines(self):
        """Generator of lines from the standard output of the process.

        Usable only if the process was started asynchronously.
        """
        if not self.started or not self.async:
            raise RuntimeError('The process has not been started.')
        if self.alive:
            stdout = self._process.stdout
        else:
            stdout = io.StringIO()
            stdout.write(self.stdout_data)
            stdout.seek(0)
        try:
            while not stdout.closed:
                try:
                    line = ''
                    self._process_lock.acquire()
                    if not stdout.closed:
                        line = stdout.readline()
                    self._process_lock.release()
                    if line == '':
                        break
                    yield line
                finally:
                    if self._process_lock.locked():
                        self._process_lock.release()
        finally:
            self._waiter_thread.join()
            if not stdout.closed:
                stdout.close()

    def __iter__(self):
        return self.readlines()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.alive:
            self.stop()





from .model import *
from .solvers import *
from .minizinc import *
from .templates import *

__all__ = [
    'Solutions', 'minizinc', 'mzn2fzn', 'solns2out', 'MiniZincError',
    'MiniZincUnsatisfiableError', 'MiniZincUnknownError',
    'MiniZincUnboundedError', 'MiniZincModel', 'Statement', 'Constraint',
    'Variable', 'ArrayVariable', 'OutputStatement', 'SolveStatement', 'Solver',
    'Gecode', 'Chuffed', 'Optimathsat', 'Opturion', 'MIPSolver', 'Gurobi',
    'CBC', 'G12Solver', 'G12Fd', 'G12Lazy', 'G12MIP', 'OscarCBLS', 'ORTools',
    'gecode', 'chuffed', 'optimathsat', 'opturion', 'gurobi', 'cbc', 'g12fd',
    'g12lazy', 'g12mip', 'oscar_cbls', 'or_tools', 'discretize', 'from_string',
    'add_package', 'add_path'
]
